from pynvml import *

import argparse, json
import datetime
import os
import logging
import torch, random
import time
import psutil
import numpy as np
from server import *
from client import *
import models, datasets

def setup_logger():
    logger = logging.getLogger("FederatedLearning")
    logger.setLevel(logging.INFO)
    handler = logging.FileHandler("fedavg_training_log.txt")
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

def log_metrics(logger, epoch, acc, loss, throughput, step_time, epoch_time, memory_usage, power_usage):
    logger.info(f"Epoch: {epoch}, Accuracy: {acc:.4f}, Loss: {loss:.4f}, Samples/sec: {throughput:.2f}, "
                f"Step time (s): {step_time:.4f}, Epoch time (s): {epoch_time:.4f}, "
                f"Memory Usage (%): {memory_usage:.2f}, Power Consumption (W): {power_usage:.2f}")
def get_gpu_power_usage():
    nvmlInit()
    handle = nvmlDeviceGetHandleByIndex(0)  # 默认使用第一个 GPU
    power = nvmlDeviceGetPowerUsage(handle) / 1000.0  # W
    nvmlShutdown()
    return power

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Federated Learning')
    parser.add_argument('-c', '--conf', dest='conf')
    args = parser.parse_args()

    with open(args.conf, 'r') as f:
        conf = json.load(f)    

    # Setup logging
    logger = setup_logger()

    # Dataset loading
    train_datasets, eval_datasets = datasets.get_dataset("./data/", conf["type"])

    # Server and clients setup
    server = Server(conf, eval_datasets)
    clients = []

    for c in range(conf["no_models"]):
        clients.append(Client(conf, server.global_model, train_datasets, c))

    print("\n\n")

    throughput_list = []  # 记录每个吞吐量
    step_time_list = []  # 记录每个 step 的时间
    epoch_time_list = []  # 记录每个 epoch 的时间
    memory_usage_list = []  # 当前 epoch 的显存使用率
    power_usage_list = []  # 当前 epoch 的功耗

    for e in range(conf["global_epochs"]):
        epoch_start_time = time.time()

        candidates = random.sample(clients, conf["k"])

        weight_accumulator = {}
        for name, params in server.global_model.state_dict().items():
            weight_accumulator[name] = torch.zeros_like(params)

        step_start_time = time.time()

        for c in candidates:
            diff = c.local_train(server.global_model)

            for name, params in server.global_model.state_dict().items():
                weight_accumulator[name].add_(diff[name])

        step_end_time = time.time()
        step_time = step_end_time - step_start_time
        step_time_list.append(step_time) # 记录 step_time

        server.model_aggregate(weight_accumulator)

        acc, loss = server.model_eval()

        # Throughput: Samples processed per second
        samples_processed = sum(len(dataset) for dataset in train_datasets)
        throughput = samples_processed / step_time # 原 samples_per_second
        throughput_list.append(throughput) # 记录 throughput 吞吐量

        epoch_end_time = time.time()
        epoch_time = epoch_end_time - epoch_start_time # 每个 epoch 的时间
        epoch_time_list.append(epoch_time) # 记录 epoch_time

        # 显存利用率 (in %)
        memory_usage = torch.cuda.memory_allocated() / (1024 ** 2) / 24 if torch.cuda.is_available() else 0
        memory_usage_list.append(memory_usage) # 当前 epoch 的显存使用率

        # 功耗
        power_usage = get_gpu_power_usage()  # Mocking power usage based on CPU utilization
        power_usage_list.append(power_usage) # 当前 epoch 的功耗

        # Log the metrics
        # log_metrics(logger, e, acc, loss, throughput, step_time, epoch_time, memory_usage, power_usage)

        print(f"Epoch {e}, acc: {acc:.4f}, loss: {loss:.4f}, "
              f"Samples/sec: {throughput:.2f}, Step time: {step_time:.4f}s, "
              f"Epoch time: {epoch_time:.4f}s, Memory Usage: {memory_usage:.2f}%, Power Usage: {power_usage:.2f}W")

    avg_epoch_time = sum(epoch_time_list) / len(epoch_time_list)
    avg_step_time = sum(step_time_list) / len(step_time_list)
    avg_throughput = sum(throughput_list) / len(throughput_list)
    avg_memory_usage = sum(memory_usage_list) / len(memory_usage_list)
    avg_power_usage = sum(power_usage_list) / len(power_usage_list)
    # 只记录平均值
    log_metrics(logger, e, acc, loss, avg_throughput, avg_step_time, avg_epoch_time, avg_memory_usage, avg_power_usage)
